; Implements a MORE$ device to which you can redirect program output.
; Different to the MORE program, no pipes are required to use it, so
; no temp files are needed. DOS has no real pipes and needs temp files
; to simulate pipes.
;
; New (more "just in time" display): tree > more$
; Old (display starts when tree is done, tempfile needed): tree | more
; Just install with DEVICE or DEVICEHIGH in your config.sys and use as
; described above then...
;
; By Eric Auer (eric -at- coli.uni-sb.de) 2003. Released under the GNU
; public license version 2 (GPL). If your DOS did not contain a copy of
; the GPL text, get one from where you downloaded this program or from
; www.gnu.org (file is about 18k in size, so including it with the MORE$
; package directly felt inappropriate: FreeDOS users have enough copies
; of GPL anyway ;-).

%define VERBOSE 1       ; give error messages on unsupported requests

	org 0   ; it is a SYS file...!!!

header:
		dd     -1       ; space for device chain pointer
		dw      8000h   ; generic char device, nothing else
		; values ("or" them) 10h-int29h 2-stdout 1-stdin...
		; DO NOT set int 29h flag, or chario.c will bypass device
		; code and use int 29h directly!
		; other bits: 800h supports open/close,
		; 2000h has "output until busy"
		; ??? 80h = is char device ???
		; (functions D/E open/close not used by FreeDOS anyway!)
		dw      Strategy
		dw      MoreIntr
		db      'MORE$   '

point           dd 0    ; pointer to parameter block
	; format: byte len, byte unit (irrelevant for non-disks),
	; byte COMMAND, word (RETURNED) STATUS, ...
	; offset 0dh: byte media descr (disks), 0eh: far pointer DATA
	; 12h: word COUNT (disk blocks / device chars)
	; 14h: word first block number (disks) (use dword below if 32bit)
	; 16h: far pointer volume id (disks) 18h: dword block number (disks)

condev          dd 0    ; pointer to CON device
conjmp          dd 0    ; temporary pointers to CON handlers

inuse           db 0    ; if 1, init is no longer possible

; if command is not 8 or 9 (those are MoreWrite), just chain on to CON ...
; however, we use int 29 for output - faster than passing on to CON

; ------------

Strategy:       ; *** not needed for MORE$ kernel patch
	mov [cs:point+2],es
	mov [cs:point],bx
	retf

; ------------

MoreIntr:       ; *** all not needed for MORE$ kernel patch
	pushf
	push ax
	push es
	push bx
	les bx,[cs:point]       ; parameter block pointer
; ???   mov byte [es:bx+17h],0          ; no error (is this used???)
				; set status word:
	and word [es:bx+3],7d00h        ; not busy, no error, error 0
	or  word [es:bx+3],100h         ; ack
	; BUG: and 7dfd / or 1 was WRONG. and 7d00 / or 100 is RIGHT.
	mov al,[es:bx+2]        ; get command code
	or al,al
	jnz further
	cmp [cs:inuse],al
	jnz eof                 ; skip init if already done
	jmp setup
eof:    pop bx
	pop es
	pop ax
	popf
	retf

; ------------

%if 0	; *** DISABLED
nosuchfunc:
	; ??? and word [es:bx+3],0fe00h ; clear ack / error code
	or word [es:bx+3],8203h ; error, busy, "unknown command"
%ifdef VERBOSE
	push bx
	push si
	; *** set SI depending on AL ...
nosuchmsg:
	mov al,[cs:si]
	or al,al
	jz nosuchtold
	int 29h ; fast DOS TTY output
nosuchtold:
	pop si
	pop bx
%endif
	jmp short eof
%else	; *** /DISABLED
%undef	VERBOSE	; *** VERBOSE is only useful in combination with
		; the above "nosuchfunc" error message output code.
%endif

; ------------

			; TODO: it would be elegant to really use the
			; "output status query" / "output until busy"

further:
	cmp al,8        ; OUTPUT
	jz writing
	cmp al,9        ; OUTPUT WITH VERIFY
	jz writing
	cmp al,10h      ; OUTPUT UNTIL BUSY (may return a smaller
	jz writing	; "count of chars" than it was called with)
	cmp al,0ah      ; OUTPUT STATUS QUERY
	jz eof		; *** simply return "not busy"
			; *** MS DOS uses this. If this returns BUSY,
			; *** DOS will wait a while before it tries
			; *** again. Only useful for serial devices!?
	cmp al,0bh      ; OUTPUT FLUSH
	jz eof		; just pretend that flush worked. Actually, we
			; never keep things "in queue", so it does ;-)

			; otherwise, CHAIN to CON device
	push si
	lds si,[cs:condev]
	mov [cs:conjmp+2],ds
	mov ax,[ds:si+6]        ; offset of Strategy call
	mov [cs:conjmp],ax
	call far [cs:conjmp]    ; a call for strategy...
	mov ax,[ds:si+8]        ; offset of Interrupt call
	mov [cs:conjmp],ax
	pop si
	pop bx
	pop es
	pop ax
	popf
	jmp far [cs:conjmp]     ; chain on to CON handler!

; ------------

writing:
	push di
	push cx
	mov cx, [es:bx+12h]     ; get count of chars to write
	les di, [es:bx+0eh]     ; get pointer to data buffer
	jmp MoreWrite


MoreNdRd0:      ; *** not needed for MORE$ kernel patch
_IOExit:
	mov ah,1        ; e.g. ah 3 done, ah 81 error al, ...
	; AL: ... 3 unknown cmd. 10 wrFault 11 rdFault 12 other fault
	les bx,[cs:point]       ; parameter block pointer
	mov [es:bx+3],ax        ; write status value (correct?)
	pop cx
	pop di
	jmp eof

; ------------

MoreWrite:      ; Input: count CX, buffer ESDI
		jcxz    MoreNdRd0               ; _IOExit if nothing to write
		call GetScreenGeometry

       ; *** check if screen "full"
MoreWrLoop:     push bx
		push cx
		push dx
		mov ah, 3                       ; get cursor position
		mov bh, 0                       ; page 0, as with int 29h
		int 10h                         ; size ch..cl, pos (dl,dh)
		mov al,[es:di]                  ; what would we print next?
		cmp al,10                       ; LF
		jnz MoreNoLF
		inc dh                          ; simulate LF
		jmp short MoreSimDone
		; Parse 9 (HT, tab)? dl=(dl&0f8h)+8... RBIL: TTY ignores it!?
MoreNoLF:       cmp al,8                        ; BackSpace
		jnz MoreNoBS
		cmp dl,0
		jz MoreSimDone
		dec dl
		jmp short MoreSimDone
MoreNoBS:       cmp al,7                        ; Bell
		jz MoreSimDone                  ; is nonprinting
		cmp al,13                       ; CR
		jnz MoreAnyChar
		mov dl,-1                       ; jump to left
MoreAnyChar:    inc dl                          ; treat anything else as char
MoreSimDone:    cmp dl,[cs:MoreScrCols]         ; past right margin?
		jb MoreNoWrap
		inc dh                          ; then wrap around
		mov dl,0
MoreNoWrap:     cmp dh,[cs:MoreScrRows]         ; past last row?
		jb MoreNoWait
	; *** checked if scrolling

		mov al,[cs:MoreCntDown]         ; then check scroll count
		or al,al                        ; gotta wait?
		jnz MoreCountOn                 ; else continue

	; *** if screen 'full' - show a message and wait for a keypress
MoreWait:       mov bx,MoreMessage              ; "--MORE--", basically
MoreTellMsg:    mov al,[cs:bx]
		inc bx
		or al,al
		jz MoreToldMsg
		int 29h                         ; Do fast output call
		jmp short MoreTellMsg
MoreToldMsg:    

; -old-         mov ah,0        ; -old-         ; wait for key
; -old-         int 16h         ; -old-         ; get key: scan AH, char AL
		call confetch   ; -new-         ; read key from CON into AL
						; -> better ^C handling and
						; idling by really using CON

		; We could process the returned char here, to provide
		; some kind of user interface...! Not done for now.

		mov al,[cs:MoreScrRows]         ; count next screen of text
		dec al                          ; ... minus a line or two!

MoreCountOn:    dec al                          ; next int 29h will scroll
		mov [cs:MoreCntDown],al         ; count lines to go until wait
MoreNoWait:     pop dx
		pop cx
		pop bx
	; *** checked if screen 'full' - waited before scrolling on then

		mov     al,[es:di]
		inc     di
		int     29h                     ; Do fast output call
		loop    MoreWrLoop              ; Loop if more to output
MoreNdRd:       jmp     _IOExit

; ------------

GetScreenGeometry:
		push ax
		push ds
		mov ax,40h
		mov ds,ax
		mov al,[ds:84h] ; EGA/better: max row number
		mov ah,[ds:4ah] ; generic: columns on screen
		pop ds
		inc al
		cmp al,25
		jnb MoreBelief1
		mov al,25       ; did not believe row count
MoreBelief1:    cmp ah,40
		jnb MoreBelief2
		mov ah,40       ; did not believe column count
MoreBelief2:    mov [cs:MoreScrRows],al
		mov [cs:MoreScrCols],ah
		pop ax
		ret

; ------------

MoreCntDown    db 24   ; number of lines until next pause message
MoreScrRows    db 25   ; number of lines for more
MoreScrCols    db 80   ; number of columns for more
MoreMessage    db 13,10,"--MORE--",13,0        ; CR, no LF at the end!

; ------------

concall:        ; call the CON device
	; See RBIL 2F0802 (driver.sys) -> table of device driver
	; command codes etc. / char: 4=input 5=peek 6=status 7=flush
	mov [cs:devcallbufferp+2],cs
	push ds
	push es
	push bx
	push si
	push ax
	mov ax,cs
	mov es,ax
	mov bx,devcallblock
		; we assume that nothing destroys this data over time!
	lds si,[cs:condev]
	mov [cs:conjmp+2],ds
	mov ax,[ds:si+6]        ; offset of Strategy call
	mov [cs:conjmp],ax
	call far [cs:conjmp]    ; a call for strategy...
	mov ax,[ds:si+8]        ; offset of Interrupt call
	mov [cs:conjmp],ax
	call far [cs:conjmp]
	pop ax
	pop si
	pop bx
	pop es
	pop ds
	ret

; ------------

confetch:       ; read char UNLESS it is ctrl C ...
	mov byte [cs:devcallblock],0eh
	mov byte [cs:devcallcommand],5  ; peek
	call concall
	mov ax,[cs:devcallstatus]
	test ax,200h                    ; busy?
	jnz connotfetched       ; if yes, nothing fetched yet, so wait
	mov al,[cs:devcallextra]
	cmp al,3                        ; ^C ?
	jz conbreaked
	jmp short condofetch            ; remove key from buffer

connotfetched:                          ; nothing in queue yet
	; we could call IDLE (int 28h) here IF we were DOS (stack...!)
	push ax
	mov ah,84h      ; "idle, network redirector may use CPU"
	int 2ah                         ; hope this works!
	pop ax
	; int 2a.84 hopefully does not imply any special DOS state!
	jmp short confetch              ; try again

condofetch:
	mov byte [cs:devcallblock],14h
	mov byte [cs:devcallcommand],4  ; read (wait for key if needed)
	call concall
	mov ax,[cs:devcallbuffer]       ; only AL really used here
	ret

conbreaked:
	mov al,10       ; LF, as we just wrote "CR LF --MORE-- CR"...
	int 29h                         ; give feedback!
	mov al,'^'
	int 29h                         ; give feedback!
	mov al,'C'
	int 29h                         ; give feedback!
	mov al,3                        ; restore "^C" key value
	ret

; ------------

devcallblock    db 14h, 0       ; size, unit (unused)
devcallcommand  db 4    ; read
devcallstatus   dw 0    ; will test 8000 NZ if error, test 200 NZ if
		; busy, test 100 NZ if done, low byte is error code
		dd 0, 0 ; reserved
devcallextra    db 0    ; command 4: drive descriptor (unused)
			; command 5: the character
devcallbufferp  dw devcallbuffer        ; (command 4)
		dw 0    ; ... CS / segment of devcallbufferp
		dw 1    ; command 4: read 1 byte
		; (dw 1st sector ...)

devcallbuffer   db 0,0,0,0      ; here we receive text from CON

; ------------

%ifdef VERBOSE
notilbusymsg    db 13,10,"MORE$: output until busy attempted",13,10,0
; --- nostatquerymsg  db 13,10,"MORE$: status query attempted",13,10,0
noflushmsg      db 13,10,"MORE$: output flush attempted",13,10,0
%endif

; ------------

setup:  mov ax,setup            ; resident part will end here
	mov [es:bx+0eh],ax      ; in init, data buffer pointer is used to...
	mov [es:bx+10h],cs      ; ... RETURN an end of resident part pointer
	inc byte [cs:inuse]     ; block access to setup in the future

	mov ah,52h              ; find "list of lists"
	int 21h                 ; es:bx is now a pointer to there...
	les bx,[es:bx+0ch]      ; pointer to current CON device
				; *** Aitor explains: This is only the
				; *** current STDIN device. The "REAL CON"
	mov [cs:condev+2],es	; *** can be found by surfing the whole
	mov [cs:condev],bx	; *** chain (List of Lists "NUL" dev is the
				; *** first device) and selecting the last
				; *** with name "CON" and flags >= 8003h.
				; *** LoL "NUL" offset differs if DOS < 3.1!
	call GetScreenGeometry  ; init counters
	mov bl,[cs:MoreScrRows]
	dec bl
	dec bl
	mov [cs:MoreCntDown],bl ; init initial countdown

	mov bx,MoreHelloMessage
hello:  mov al,[cs:bx]  ; fetch initial hello message and show it
	or al,al
	jnz hello2
	jmp eof
hello2: int 29h         ; Do fast output call
	inc bx
	jmp short hello

MoreHelloMessage:
	db "MORE$ driver by Eric Auer 2003 loaded. Use: ... > MORE$"
	db 13,10,0

; ------------

	; request header:
	; db headersize, db subunitnumber, db COMMAND, dw STATUS
	; dd reserved, dd reserved, [0d]... differ per command:
	; ->
	; 0 - init: db unitcount, dd pointer to first byte past
	;   driver, must be smaller on return than on call if DOS 5.0+,
	;   dd commandline (call) / BPB array (return, block devices)
	;   db drive number of first unit (0=A:) dw error message flag
	;   (DOS 5.0+, set to 1 if error msg should be displayed)
	; 1 - block devices: media check: db descriptor db returned
	;   status (0=unknown 1=notchanged -1=changed)
	; 2 - block devices: build bpb: db descriptor, dd pointer to
	;   scratch sector (non-IBM formats) or first FAT sector,
	;   dd pointer to BPB (set by driver)
	; 4 - input: db descriptor (only used for block devices),
	;   dd pointer to buffer, dw bytes/sectors (char/block devices)
	;   count, dw 1st sector (block devices), dd pointer to volume
	;   id in case of error 0f, [1a] dd 1st sector (if > 0ffffh)...
	;   DR DOS 6 simply has [14] dd for 1st sector right away...
	; 5 - char devices: peek input: db "byte from device if BUSY
	;   flag is not set"
	; 6 - char devices: get input status (no further data at [0d])
	; 7 - char devices: flush input buffers (no further data)
	; 8 - output (like command 4)
	; 9 - output with verify (like command 4)
	; a - char devices: get output status (no further data)
	; b - char devices: flush output buffers (no further data)
	; d - open device (optional) (no further data)
	; e - close device (optional) (no further data)
	; 10 - char devices: output until busy: db (unused), dd pointer
	;   to buffer, dw number of bytes to write (call) / number of
	;   bytes actually written (return).
	; ...
	; cd-rom: 80 - read long 82 - read long prefetch: db mode
	;   (0=HSG, 1=RedBook), dd pointer to buffer (unused for
	;   command 82), dw sector count (for command 82, 0 means
	;   advisory seek, other means must seek), dd 1st sector
	;   number (HSG: 4500*m + 75*s + frame - 150, RedBook: bytes
	;   for frame, second, minute, unused), db read mode (0:
	;   cooked, 2k/frame, 1: raw, 2352by/frame), db interleave
	;   size (sectors-in-a-row size), [1a] db interleave skip
	;   (how far to skip between consecutive portions)
	; cd-rom: 83 - seek: db mode (HSG/RB), dd unused (buffer
	;   pointer), dw unused (count), [14] dd 1st sector
	; cd-rom: other commands available for AUDIO...
	; ...

	; request status word: bit15=error bit9=busy bit8=done
	; bit0..7=error code if bit 15 set: 0 write protect,
	; 1 unknown unit, 2 drive not ready, 3 unknown command,
	; a write error, b read error, 8 sector not found, ...
	; e no CD-ROM in drive, 7 unknown media, 6 seek error ...

